from __future__ import division, print_function, absolute_import, unicode_literals

import os
from javactl.util import CaseClass, normalize_path, omap, oget


class AppSetting(CaseClass):
    def __init__(self, name=None, home=None, jar=None, entry_point=None, command=None, pid_file=None):
        # constraints
        assert name is not None, 'app.name is required'
        assert home is not None, 'app.home is required'
        assert os.path.isabs(home), 'app.home must be an absolute path'
        assert (jar is None) != (command is None), 'Either app.jar or app.command but not both must be given'
        assert jar is not None or entry_point is None, 'app.entry_point must be used with app.jar'

        normalize = lambda p: normalize_path(p, home)

        CaseClass.__init__(
            self,
            ('name', name),
            ('home', home),
            ('jar', omap(normalize, jar)),
            ('entry_point', entry_point),
            ('command', omap(normalize, command)),
            ('pid_file', omap(normalize, pid_file))
        )

    def is_duplicate_allowed(self):
        return self.pid_file is not None

    def get_args(self, java_args):
        if self.jar is not None:
            if self.entry_point is not None:
                return java_args + ['-cp', self.jar, self.entry_point]
            else:
                return java_args + ['-jar', self.jar]
        else:
            return [self.command]



import os
import threading


# steeled from http://codereview.stackexchange.com/questions/6567/
class LoggerWrapper(threading.Thread):
    """
    Read text message from a pipe and redirect them
    to a logger (see python's logger module),
    the object itself is able to supply a file
    descriptor to be used for writing

    fdWrite ==> fdRead ==> pipeReader
    """

    def __init__(self, logger, level):
        """
        Setup the object with a logger and a loglevel
        and start the thread
        """

        # Initialize the superclass
        threading.Thread.__init__(self)

        # Make the thread a Daemon Thread (program will exit when only daemon
        # threads are alive)
        self.daemon = True

        # Set the logger object where messages will be redirected
        self.logger = logger

        # Set the log level
        self.level = level

        # Create the pipe and store read and write file descriptors
        self.fdRead, self.fdWrite = os.pipe()

        # Create a file-like wrapper around the read file descriptor
        # of the pipe, this has been done to simplify read operations
        self.pipeReader = os.fdopen(self.fdRead)

        # Start the thread
        self.start()

    def fileno(self):
        """
        Return the write file descriptor of the pipe
        """
        return self.fdWrite

    def run(self):
        """
        This is the method executed by the thread, it
        simply read from the pipe (using a file-like
        wrapper) and write the text to log.
        NB the trailing newline character of the string
           read from the pipe is removed
        """

        # Endless loop, the method will exit this loop only
        # when the pipe is close that is when a call to
        # self.pipeReader.readline() returns an empty string
        while True:

            # Read a line of text from the pipe
            messageFromPipe = self.pipeReader.readline()

            # If the line read is empty the pipe has been
            # closed, do a cleanup and exit
            # WARNING: I don't know if this method is correct,
            # further study needed
            if len(messageFromPipe) == 0:
                self.pipeReader.close()
                os.close(self.fdRead)
                return
            # end if

            # Remove the trailing newline character frm the string
            # before sending it to the logger
            if messageFromPipe[-1] == os.linesep:
                messageToLog = messageFromPipe[:-1]
            else:
                messageToLog = messageFromPipe

            # Send the text to the logger
            self._write(messageToLog)

    def _write(self, message):
        """
        Utility method to send the message
        to the logger with the correct loglevel
        """
        self.logger.log(self.level, message)



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import subprocess
from javactl.util import CaseClass, normalize_path, omap, oget


class JavaSetting(CaseClass):
    class Memory(CaseClass):
        def __init__(self,
                     java_version,  # just use for assertion
                     heap_min=None,
                     heap_max=None,
                     perm_min=None,
                     perm_max=None,
                     metaspace_min=None,
                     metaspace_max=None,
                     new_min=None,
                     new_max=None,
                     survivor_ratio=None,
                     target_survivor_ratio=None):
            # constraints
            assert perm_min is None or java_version < 1.8, 'java.memory.perm_min is not applicable to java >= 1.8'
            assert perm_max is None or java_version < 1.8, 'java.memory.perm_man is not applicable to java >= 1.8'
            assert metaspace_min is None or java_version >= 1.8, \
                'java.memory.metaspace_min is not applicable to java < 1.8'
            assert metaspace_max is None or java_version >= 1.8, \
                'java.memory.metaspace_max is not applicable to java < 1.8'

            CaseClass.__init__(
                self,
                ('heap_min', heap_min),
                ('heap_max', heap_max),
                ('perm_min', perm_min),
                ('perm_max', perm_max),
                ('metaspace_min', metaspace_min),
                ('metaspace_max', metaspace_max),
                ('new_min', new_min),
                ('new_max', new_max),
                ('survivor_ratio', survivor_ratio),
                ('target_survivor_ratio', target_survivor_ratio)
            )

        def get_opts(self):
            xs = [
                omap(lambda s: '-Xms%s' % s, self.heap_min),
                omap(lambda s: '-Xmx%s' % s, self.heap_max),
                omap(lambda s: '-XX:PermSize=%s' % s, self.perm_min),
                omap(lambda s: '-XX:MaxPermSize=%s' % s, self.perm_max),
                omap(lambda s: '-XX:MetaspaceSize=%s' % s, self.metaspace_min),
                omap(lambda s: '-XX:MaxMetaspaceSize=%s' % s, self.metaspace_max),
                omap(lambda s: '-Xmn%s' % s, self.new_min),
                omap(lambda s: '-XX:MaxNewSize=%s' % s, self.new_max),
                omap(lambda x: '-XX:SurvivorRatio=%d' % x, self.survivor_ratio),
                omap(lambda x: '-XX:TargetSurvivorRatio=%d' % x, self.target_survivor_ratio),
            ]
            return [x for x in xs if x is not None]

    class JMX(CaseClass):
        def __init__(self, port=None, ssl=None, authenticate=None):
            CaseClass.__init__(self, ('port', port), ('ssl', ssl), ('authenticate', authenticate))

        def get_opts(self):
            if self.port is None:
                return []
            xs = [
                '-Dcom.sun.management.jmxremote',
                omap(lambda x: '-Dcom.sun.management.jmxremote.port=%d' % x, self.port),
                omap(lambda b: '-Dcom.sun.management.jmxremote.ssl=%s' % str(b).lower(), self.ssl),
                omap(lambda b: '-Dcom.sun.management.jmxremote.authenticate=%s' % str(b).lower(), self.authenticate),
            ]
            return [x for x in xs if x is not None]

    def __init__(self,
                 home=None,
                 version=None,
                 server=None,
                 memory=None,
                 jmx=None,
                 env=None,
                 option=None):
        # constraints
        assert home is not None and os.path.isabs(home), 'java.home is required and must be an absolute path'
        assert version is not None, 'java.version is required'

        # TODO: check value types and format
        assert env is None or isinstance(env, dict), 'java.env must be a dict'
        assert option is None or isinstance(option, list), 'java.env must be a list'

        CaseClass.__init__(
            self,
            ('home', home),
            ('version', version),
            ('server', server),
            ('memory', JavaSetting.Memory(version, **oget(memory, {}))),
            ('jmx', JavaSetting.JMX(**oget(jmx, {}))),
            ('env', oget(env, {})),
            ('option', oget(option, [])),
        )

    def get_executable(self):
        return os.path.join(self.home, 'bin', 'java')

    def get_opts(self):
        sv = ['-server'] if self.server else []
        ev = ['-D%s=%s' % (k, v) for k, v in self.env.items()]
        return sv + self.memory.get_opts() + self.jmx.get_opts() + ev + self.option

    def get_args(self):
        return [self.get_executable()] + self.get_opts()



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import errno
import subprocess
import six


#
# Path operation
#
def normalize_path(path, base_dir):
    return path if os.path.isabs(path) else os.path.join(base_dir, path)


#
# Process operation
#
def execute_command(args, shell=False, cwd=None, env=None, stdin=None, stdout=None, stderr=None):
    """
    Execute external command

    :param args: command line arguments : [string]
    :param shell: True when using shell : boolean
    :param cwd: working directory : string
    :param env: environment variables : dict
    :param stdin: standard input
    :param stdout: standard output
    :param stderr: standard error
    :return: return code
    """
    return subprocess.call(args=args, shell=shell, cwd=cwd, env=dict(os.environ, **(oget(env, {}))),
                           stdin=stdin, stdout=stdout, stderr=stderr)


def capture_command(args, shell=False, cwd=None, env=None, stdin=None):
    """
    Execute external command and capture output

    :param args: command line arguments : [string]
    :param shell: True when using shell : boolean
    :param cwd: working directory : string
    :param env: environment variables : dict
    :param stdin: standard input
    :return: tuple of return code, stdout data and stderr data
    """
    p = subprocess.Popen(
        args, shell=shell, cwd=cwd, env=dict(os.environ, **(oget(env, {}))),
        stdin=stdin, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout_data, stderr_data = p.communicate()
    return p.returncode, stdout_data, stderr_data


def execute_command_with_pid(args, pid_file=None, shell=False, cwd=None, env=None,
                             stdin=None, stdout=None, stderr=None):
    if pid_file is None:
        return execute_command(args, shell, cwd, env, stdin, stdout, stderr)
    else:
        try:
            p = subprocess.Popen(
                args, shell=shell, cwd=cwd, env=dict(os.environ, **(oget(env, {}))),
                stdin=stdin, stdout=stdout, stderr=stderr)
            with open(pid_file, 'w') as f:
                f.write(str(p.pid))
            ret = p.wait()
        finally:
            # clean up pid file
            if pid_file is not None and os.path.exists(pid_file):
                os.remove(pid_file)
        return ret


def pid_exists(pid):
    # Steeled from https://github.com/giampaolo/psutil/blob/master/psutil/_psposix.py
    """Check whether pid exists in the current process table."""
    if pid == 0:
        # According to "man 2 kill" PID 0 has a special meaning:
        # it refers to <<every process in the process group of the
        # calling process>> so we don't want to go any further.
        # If we get here it means this UNIX platform *does* have
        # a process with id 0.
        return True
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH) therefore we should never get
            # here. If we do let's be explicit in considering this
            # an error.
            raise err
    else:
        return True


#
# Functional operation
#
def omap(function, optional):
    """Map optional value"""
    return None if optional is None else function(optional)


def oget(optional, default=None):
    return default if optional is None else optional


#
# String util
#
def is_unicode(s):
    return (six.PY2 and isinstance(s, unicode)) or (six.PY3 and isinstance(s, str))


def is_strlike(s):
    return isinstance(s, (six.string_types, bytes))


def to_unicode(s, encoding=None, errors='strict'):
    """
    Make unicode string from any value
    :param s:
    :param encoding:
    :param errors:
    :return: unicode
    """
    encoding = encoding or 'utf-8'

    if is_unicode(s):
        return s
    elif is_strlike(s):
        return s.decode(encoding, errors)
    else:
        if six.PY2:
            return str(s).decode(encoding, errors)
        else:
            return str(s)


def to_str(s, encoding=None, errors='strict'):
    """
    Make str from any value
    :param s:
    :param encoding:
    :param errors:
    :return: str (not unicode in Python2, nor bytes in Python3)
    """
    encoding = encoding or 'utf-8'

    if is_strlike(s):
        if six.PY2:
            return s.encode(encoding, errors) if isinstance(s, unicode) else s
        else:
            return s.decode(encoding, errors) if isinstance(s, bytes) else s
    else:
        return str(s)



from javactl.util.case_class import CaseClass

import six


class DataSize(CaseClass):
    _scales = {'k': 2 ** 10, 'm': 2 ** 20, 'g': 2 ** 30}

    def __init__(self, size):
        assert isinstance(size, int) or isinstance(size, six.string_types)

        if isinstance(size, int) or size.isdigit():
            value = int(size)
            scale = ''
        else:
            val, last = size[:-1], size[-1:]
            assert last.lower() in self._scales.keys(), 'Unknown scale: %s' % last
            assert val.isdigit(), 'Size must be an integer: %s' % val

            value = int(val)
            scale = last
        CaseClass.__init__(self, ('value', value), ('scale', scale))

    def bytes(self):
        return self.value * self._scales.get(self.scale.lower(), 1)

    def __str__(self):
        return '%d%s' % (self.value, self.scale)



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
try:
    import syslog
    SYSLOG_AVAILABLE = True
except ImportError:
    SYSLOG_AVAILABLE = False
from javactl.logger.logger import Logger
from javactl.util import util


class SystemLogger(Logger):
    def __init__(self, dry_run=False):
        self.name = os.path.basename(sys.argv[0])
        self.dry_run = dry_run
        super(SystemLogger, self).__init__('SystemLogger[%s]' % self.name)

    def _log(self, priority, message):
        if SYSLOG_AVAILABLE:
            if self.dry_run:
                print('Would write to syslog: priority=%s, message=%s' % (priority, message))
            else:
                syslog.openlog(self.name, syslog.LOG_PID)
                syslog.syslog(priority, util.to_str(message, 'utf-8', 'ignore'))
                syslog.closelog()



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import traceback
from datetime import datetime
from javactl.setting.setting import Setting
from javactl.logger import SystemLogger
from javactl.executor.executor import Executor
from javactl.exceptions import DuplicateError


def main():
    """
    Main function
    """

    return_code = 0
    try:
        setting = Setting().parse_args(sys.argv).load_config()
        now = datetime.now()
        ex = Executor(setting, SystemLogger(setting.dry_run))
        result = ex.check_requirement().create_directories().clean_old_logs(now).execute(now)
        if result.failed:
            return_code = 1
    except DuplicateError:
        return 0
    except Exception as e:
        print('%s: %s' % (e.__class__.__name__, e))
        # print(traceback.print_exc())
        return 2
    return return_code



from __future__ import division, print_function, absolute_import, unicode_literals

import logging
from logging.handlers import RotatingFileHandler
from javactl.logger.logger_wrapper import LoggerWrapper
from javactl.util import util


def get_console_logger(path, max_bytes, backup_count, encoding='utf-8'):
    logger = logging.getLogger('ConsoleLogger')
    logger.setLevel(logging.INFO)

    # add a handler
    handler = RotatingFileHandler(path, 'a', max_bytes, backup_count)
    handler.setFormatter(logging.Formatter(util.to_str('[%(asctime)s] %(message)s', encoding, 'ignore')))
    logger.addHandler(handler)

    return LoggerWrapper(logger, logging.INFO)



from __future__ import division, print_function, absolute_import, unicode_literals

import subprocess
from itertools import chain
import yaml
import six
from javactl.setting import arg_parser
from javactl.setting.app_setting import AppSetting
from javactl.setting.java_setting import JavaSetting
from javactl.setting.log_setting import LogSetting
from javactl.setting.os_setting import OSSetting
from javactl.util import CaseClass, omap, oget


class Setting(CaseClass):
    """Manages all settings."""

    def __init__(self,
                 config_path=None,
                 extra_args=None,
                 dry_run=False,
                 app_setting=None,
                 java_setting=None,
                 log_setting=None,
                 os_setting=None,
                 pre_commands=None,
                 post_commands=None):
        """
        :param config_path:
        :param extra_args: arguments for Java application
        :param dry_run:
        :param app_setting:
        :param java_setting:
        :param log_setting:
        :param os_setting:
        :param pre_commands:
        :param post_commands:
        :return:
        """
        CaseClass.__init__(
            self,
            ('config_path', config_path),
            ('extra_args', oget(extra_args, [])),
            ('dry_run', dry_run),
            ('app_setting', app_setting),
            ('java_setting', java_setting),
            ('log_setting', log_setting),
            ('os_setting', os_setting),
            ('pre_commands', oget(pre_commands, [])),
            ('post_commands', oget(post_commands, []))
        )

    def copy(self, **args):
        d = self.values()
        d.update(args)
        return Setting(**d)

    def parse_args(self, argv):
        option, config_path, extra_args = arg_parser.parser.parse_args(argv[1:])
        return self.copy(config_path=config_path, extra_args=extra_args, dry_run=option.dry_run)

    def load_config(self):
        if not self.config_path:
            return self

        with open(self.config_path, 'rb') as f:
            data = yaml.load(f.read().decode('utf-8'))

        app_setting = AppSetting(**data.get('app', {}))
        java_setting = JavaSetting(**data.get('java', {}))
        log_setting = LogSetting(app_setting.home, **data.get('log', {}))
        os_setting = OSSetting(**data.get('os', {}))

        pre_commands = data.get('pre', [])
        post_commands = data.get('post', [])

        for commands in [pre_commands, post_commands]:
            assert isinstance(commands, list), 'pre/post must be lists or empty'
            assert all(isinstance(s, six.string_types) for s in commands), 'each element of pre/post must be a string'

        return self.copy(app_setting=app_setting, java_setting=java_setting, log_setting=log_setting,
                         os_setting=os_setting, pre_commands=pre_commands, post_commands=post_commands)

    def get_args(self, now):
        return list(chain.from_iterable([
            self.app_setting.get_args(self.java_setting.get_args() + self.log_setting.get_opts(now)),
            self.extra_args,
        ]))

    def get_environ(self, now):
        d = {
            'JAVA_HOME': self.java_setting.home,
            'JAVA_OPTS': subprocess.list2cmdline(self.java_setting.get_opts() + self.log_setting.get_opts(now))
        }
        d.update(self.os_setting.env)
        return d



from __future__ import division, print_function, absolute_import, unicode_literals

from optparse import OptionParser
from javactl.util import oget

VERSION = 'javactl %s' % __import__('javactl').__version__

USAGE = """%prog [options...] <config_path> [args...]"""


class Parser(OptionParser):
    def parse_args(self, args=None, values=None):
        """
        Parse manually since args can contain arguments for Java application.
        :param args: [string]
        :return: (option : dict, config_path : string, args : [string])
        """

        # find the first non-optional arg
        args = oget(args, [])

        i = 0
        for i in range(len(args)):
            if not args[i].startswith('-'):
                break
        ls, rs = args[:i + 1], args[i + 1:]

        option, largs = OptionParser.parse_args(self, oget(ls, []))
        rargs = largs + rs

        if not rargs:
            self.print_help()
            self.exit(2)
            return

        return option, rargs[0], rargs[1:]


def _get_parser():
    p = Parser(usage=USAGE, version=VERSION)

    p.add_option(
        '--check', action='store_true', dest='dry_run', default=False, help='dry run mode'
    )

    return p


parser = _get_parser()



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import os
import re
import time
import getpass
import glob
import subprocess
from datetime import datetime, timedelta
from javactl.util import execute_command, capture_command, execute_command_with_pid, pid_exists, omap, oget
from javactl.logger.console_logger import get_console_logger
from javactl.exceptions import DuplicateError


class Executor(object):
    def __init__(self, setting, logger, failed=False):
        self.setting = setting
        self.logger = logger
        self.failed = failed

    def copy(self, **args):
        d = {'setting': self.setting, 'logger': self.logger, 'failed': self.failed}
        d.update(args)
        return Executor(**d)

    def check_requirement(self):
        return self._check_user()._check_java_version()._check_duplicate()

    def _check_user(self):
        actual = getpass.getuser()
        expect = self.setting.os_setting.user
        assert actual == expect, "This application must be run as '%s', but you are '%s'." % (expect, actual)
        return self

    def _check_java_version(self):
        ret, stdout, stderr = capture_command([self.setting.java_setting.get_executable(), '-version'])
        assert ret == 0, 'Failed to get java version: %s' % stderr

        first_line = ''.join(stderr.splitlines()[:1])
        m = re.compile(r"""java version \"(\d+[.]\d+)[.]\d+_\d+\"""").match(first_line)
        if m:
            actual = float(m.group(1))
        else:
            raise AssertionError('Unexpected java version: %s' % first_line)

        expect = self.setting.java_setting.version
        assert actual == expect, "Unexpected Java version: expect='%s', actual='%s'." % (expect, actual)
        return self

    def _check_duplicate(self):
        p = self.setting.app_setting.pid_file
        if p is not None:
            if os.path.exists(self.setting.app_setting.pid_file):
                with open(p) as f:
                    pid = int(f.read())
                if pid_exists(pid):
                    self.logger.info('%s is already running. Skipped: config=%s' % (
                        self.setting.app_setting.name, self.setting.config_path))
                    raise DuplicateError
        return self

    def create_directories(self):
        dirs = [d for d in [
            omap(os.path.dirname, self.setting.log_setting.console.prefix),
            omap(os.path.dirname, self.setting.log_setting.gc.prefix),
            self.setting.log_setting.dump.prefix,
            omap(os.path.dirname, self.setting.log_setting.error.path),
        ] if d is not None]

        for d in dirs:
            if not os.path.exists(d):
                self._create_directory(d)
        return self

    # OS operations
    def _create_directory(self, path):
        if self.setting.dry_run:
            print('Would create directory: %s' % path)
        else:
            print('Creating directory: %s' % path)
            os.makedirs(path)

    def _delete_file(self, path):
        if self.setting.dry_run:
            print('Would delete file: %s' % path)
        else:
            print('Deleting file: %s' % path)
            os.remove(path)

    def clean_old_logs(self, now):
        for setting in [self.setting.log_setting.console, self.setting.log_setting.gc]:
            if setting.prefix is None or setting.preserve is None:
                continue
            for path in glob.iglob(setting.prefix + '*'):
                mtime = datetime.fromtimestamp(os.path.getmtime(path))
                if mtime < now - timedelta(days=setting.preserve):
                    self._delete_file(path)
        return self

    def execute(self, now):
        return self._execute_commands(self.setting.pre_commands, now)._execute_application(now)._execute_commands(
            self.setting.post_commands, now)

    def _execute_application(self, now):
        if self.failed:
            return self

        failed = False
        self.logger.info(
            '%s started: config=%s, args=%s' % (
                self.setting.app_setting.name, self.setting.config_path, self.setting.extra_args))

        time_start = time.time()
        if self.setting.dry_run:
            print('Would execute: cwd=%s, cmd=[\n%s\n], env={\n%s\n}, output=%s' % (
                self.setting.app_setting.home,
                '\n'.join('  %s' % s for s in self.setting.get_args(now)),
                '\n'.join('  %s: %s' % kv for kv in self.setting.get_environ(now).items()),
                oget(self.setting.log_setting.console.get_path(now), 'stdout')
            ))
            ret = 0
        else:
            out_path = self.setting.log_setting.console.get_path(now)
            if out_path is None:
                stdout = sys.stdout
                stderr = sys.stderr
            else:
                stdout = get_console_logger(
                    out_path,
                    self.setting.log_setting.console.max_size.bytes(),
                    self.setting.log_setting.console.backup)
                stderr = stdout
            ret = execute_command_with_pid(
                self.setting.get_args(now),
                self.setting.app_setting.pid_file,
                shell=False,
                cwd=self.setting.app_setting.home,
                env=dict(os.environ, **(oget(self.setting.get_environ(now), {}))),
                stdin=sys.stdin, stdout=stdout, stderr=stderr)

        elapsed = time.time() - time_start

        if ret == 0:
            self.logger.info('%s ended successfully: elapsed=%ds' % (self.setting.app_setting.name, elapsed))
        else:
            self.logger.error(
                '%s ended with error: return_code=%d, elapsed=%ds' % (self.setting.app_setting.name, ret, elapsed))
            failed = True
        return self.copy(failed=failed)

    def _execute_commands(self, commands, now):
        failed = self.failed
        for cmd in commands:
            if self.setting.dry_run:
                print('Would execute: %s' % cmd)
            else:
                ret = execute_command(cmd, shell=True, cwd=self.setting.app_setting.home,
                                      env=self.setting.get_environ(now))
                if ret != 0:
                    self.logger.error('Failed to execute: app=%s, cmd=%s, return_code=%d' % (
                        self.setting.app_setting.name, cmd, ret))
                    failed = True
        return self.copy(failed=failed)



from __future__ import division, print_function, absolute_import, unicode_literals


class JavaCtlError(Exception):
    """Base class of application specific exceptions"""


class DuplicateError(JavaCtlError):
    """Application is already running."""



from __future__ import division, print_function, absolute_import, unicode_literals

import six
from javactl.util import CaseClass, normalize_path, omap, oget, DataSize


class OSSetting(CaseClass):
    def __init__(self, user=None, env=None):
        # constraints
        assert user is not None, 'os.user is required'
        assert isinstance(user, six.string_types), 'os.user must be a string'
        assert env is None or isinstance(env, dict), 'os.env must be a dict'

        d = dict((k, str(v)) for k, v in oget(env, {}).items())
        CaseClass.__init__(self, ('user', user), ('env', d))



from __future__ import division, print_function, absolute_import, unicode_literals

from javactl.util import CaseClass, normalize_path, omap, oget, DataSize


class LogSetting(CaseClass):
    class Console(CaseClass):
        def __init__(self, home, prefix=None, max_size=None, backup=None, preserve=None):
            # constraints
            assert backup is None or isinstance(backup, int), 'log.console.backup must be an integer'
            assert preserve is None or isinstance(backup, int), 'log.console.preserve must be an integer'

            CaseClass.__init__(
                self,
                ('prefix', omap(lambda p: normalize_path(p, home), prefix)),
                ('max_size', omap(DataSize, max_size)),
                ('backup', backup),
                ('preserve', preserve))

        def get_path(self, now):
            return omap(lambda p: p + now.strftime('_%Y%m%d_%H%M%S.log'), self.prefix)

    class GC(CaseClass):
        def __init__(self, home, prefix=None, max_size=None, backup=None, preserve=None):
            # constraints
            assert backup is None or isinstance(backup, int), 'log.gc.backup must be an integer'
            assert preserve is None or isinstance(backup, int), 'log.gc.preserve must be an integer'

            CaseClass.__init__(
                self,
                ('prefix', omap(lambda p: normalize_path(p, home), prefix)),
                ('max_size', omap(DataSize, max_size)),
                ('backup', backup),
                ('preserve', preserve))

        def get_path(self, now):
            return omap(lambda p: p + now.strftime('_%Y%m%d_%H%M%S.log'), self.prefix)

        def get_opts(self, now):
            xs = []
            if self.prefix is not None:
                xs.extend([
                    '-verbose:gc',
                    '-XX:+PrintGCDateStamps',
                    '-XX:+PrintGCDetails',
                    '-Xloggc:%s' % self.get_path(now),
                ])
                if self.max_size is not None:
                    xs.extend([
                        '-XX:+UseGCLogFileRotation',
                        '-XX:GCLogFileSize=%s' % self.max_size])
                    if self.backup is not None:
                        xs.append('-XX:NumberOfGCLogFiles=%s' % self.backup)
            return xs

    class Dump(CaseClass):
        def __init__(self, home, prefix=None):
            CaseClass.__init__(self, ('prefix', omap(lambda p: normalize_path(p, home), prefix)))

        def get_opts(self):
            return [] if self.prefix is None else [
                '-XX:+HeapDumpOnOutOfMemoryError',
                '-XX:HeapDumpPath=%s' % self.prefix,
            ]

    class Error(CaseClass):
        def __init__(self, home, path=None):
            CaseClass.__init__(self, ('path', omap(lambda p: normalize_path(p, home), path)))

        def get_opts(self):
            return [] if self.path is None else [
                '-XX:ErrorFile=%s' % self.path,
            ]

    def __init__(self, home, console=None, gc=None, dump=None, error=None):
        CaseClass.__init__(
            self,
            ('console', LogSetting.Console(home, **oget(console, {}))),
            ('gc', LogSetting.GC(home, **oget(gc, {}))),
            ('dump', LogSetting.Dump(home, **oget(dump, {}))),
            ('error', LogSetting.Error(home, **oget(error, {}))),
        )

    def get_opts(self, now):
        return self.gc.get_opts(now) + self.dump.get_opts() + self.error.get_opts()
